/*
    Copyright (c) 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __DATABASE_H__
#define __DATABASE_H__

#include <boost/thread/mutex.hpp>
#include <boost/shared_ptr.hpp>
#include <map>
#include <string>
#include <list>

#include "sqlite3_wrapper.h"
#include "tasks.h"

#define TRANSMISSION_QUALITY_UPSTREAM_MAX   62 // best
#define TRANSMISSION_QUALITY_DOWNSTREAM_MAX 6  // worst

namespace dSA
{

class Database
{
public:
    typedef std::pair<std::string, int> field;
    typedef std::map<std::string, boost::shared_ptr<field> > result_data;

    typedef struct
    {
        boost::shared_ptr<dSLongTask> task;
        boost::shared_ptr<result_data> data;
    } task_result_summary;

    typedef struct
    {
        std::string dsid;
        int upstream;
        int downstream;
        time_t timestamp;
        int errorcode;
    } transmission_quality_device_record;

    Database();
    static Database *createInstance();
    static Database *getInstance();
    ~Database();

    /// \brief Retrieve a value from the internal configuration dable
    std::string getConfig(std::string key);

    /// \brief Saves a task in the database
    /// \parameter task task that should be saved
    /// \return task ID
    long long int saveTask(boost::shared_ptr<const dSLongTask> task);

    /// \brief Removes a task given by id from the database
    void removeTask(long long int task_id);

    /// \brief Remove tasks given by status from the database
    void removeTasksByStatus(task_status status);

    /// \brief Update the status of a task in the databse
    ///
    /// This function will automatically take care of satting the relevant
    /// timestamps.
    void updateTaskStatus(long long int task_id, task_status status);

    /// \brief Set new status on all tasks with given old status
    void updateTaskStatus(task_status old_status, task_status new_status);

    /// \brief Load task type information
    task_type getTaskType(long long int task_id);

    /// \brief Load task from the database given by id
    boost::shared_ptr<dSLongTask> getTask(long long int task_id);

    /// \brief Retrieves a list of tasks that have a certain state
    boost::shared_ptr<std::list<boost::shared_ptr<dSLongTask> > >
                                          getTasksByStatus(task_status status);

    void saveTransmissionQuality(long long int task_id, std::string device_dsid,
                                 uint16_t upstream, uint8_t downstream,
                                 int error_code);

    /// \brief Returns the summarized version of task data.
    ///
    /// The data and the keys will vary depending on the task
    boost::shared_ptr<task_result_summary>
                                    getTaskResultSummary(long long int task_id);

    boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
                                        loadTasks(long long int limit,
                                                  long long int offset);

    long long int transmissionQualityResultCount(long long int task_id);
    boost::shared_ptr<std::list<
        boost::shared_ptr<const transmission_quality_device_record> > >
            getTransmissionQualityResults(long long int task_id,
                                          long long int limit,
                                          long long int offset);
private:
    static Database* m_instance;
    SQLite3 *m_sqlite3;
    boost::mutex m_mutex;
    int m_version;

    boost::shared_ptr<dSLongTask> createTaskFromRow(
                                    boost::shared_ptr<SQLite3::row_result> rr);
    boost::shared_ptr<field> singleValueFromQuery(
                                    boost::shared_ptr<SQLite3::query_result> q);
    boost::shared_ptr<transmission_quality_device_record>
                  TQDevRecordFromRow(boost::shared_ptr<SQLite3::row_result> rr);
};

} // namespace

#endif//__DATABASE_H__
