/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#include <boost/pointer_cast.hpp>
#include <errno.h>
#include <string.h>
#include <exception>
#include <iomanip>

#include "log.h"
#include "util.h"
#include "exceptions.h"
#include "config_manager.h"
#include "hwdesc.h"
#include "storage.h"
#include "ds-assistant.h"
#include "database.h"
#include "traffic-decode.h"

#define INVALID_TASK_ID 0

namespace dSA
{

dS_Assistant* dS_Assistant::m_instance = NULL;

dS_Assistant* dS_Assistant::createInstance()
{
    if (m_instance == NULL)
    {
        m_instance = new dS_Assistant();
        if (m_instance == NULL)
        {
            throw _Exception("could not create dS Assistant class");
        }
    }

    return m_instance;
}

dS_Assistant* dS_Assistant::getInstance()
{
    if (m_instance == NULL)
    {
        throw _Exception("dS Assistant not initialized");
    }

    return m_instance;
}
dS_Assistant::dS_Assistant() : m_shutdown_flag(false),
                               m_running(false),
                               m_task_thread(0),
                               m_long_task_thread(0)
{
    int ret;

    ConfigManager *cfg = ConfigManager::getInstance();
    std::string logdir(cfg->getOption(CFG_LOG_DIR));
    std::string logfile(logdir +  + "/dsa.log");
    m_logfile = fopen(logfile.c_str(), "a");
    if (NULL == m_logfile)
    {
        throw _Exception("failed to open logfile \'" + logfile + "\': " +
                         std::string(strerror(errno)));
    }

    Output2DSID::SetLogdirectory(logdir);
    DSIDLog::ReportingLevel() =
        FILELog::FromString((cfg->getOption(CFG_LOG_LEVEL)).c_str());

    Output2FILE::SetStream(m_logfile);
    FILELog::ReportingLevel() =
        FILELog::FromString((cfg->getOption(CFG_LOG_LEVEL)).c_str());
    FILE_LOG(logINFO) << "dS Assistant startup";

    SetNullId(m_own_dsid);

    ret = pthread_mutex_init(&m_task_list_mutex, NULL);
    if (ret != 0)
    {
        throw _Exception("failed to initialize task list mutex: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_mutex_init(&m_long_task_list_mutex, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        throw _Exception("failed to initialize long task list mutex: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_cond_init(&m_wakeup_task_cond, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        pthread_mutex_destroy(&m_long_task_list_mutex);
        throw _Exception("failed to initialize task thread wakeup condition: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_cond_init(&m_wakeup_long_task_cond, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        pthread_mutex_destroy(&m_long_task_list_mutex);
        pthread_cond_destroy(&m_wakeup_task_cond);
        throw _Exception("failed to initialize task thread wakeup condition: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::connect()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);

    if (m_dsm_api_wrapper != NULL)
    {
        throw _Exception("connection already exists");
    }

    ConfigManager *cfg = ConfigManager::getInstance();
    m_dsm_api_wrapper = boost::shared_ptr<dSMApiWrapper>(
                    new dSMApiWrapper(cfg->getOption(CFG_DS485_DAEMON_URL)));

    try
    {
        m_dsm_api_wrapper->getOwnDSID(&m_own_dsid);
    }
    catch (Exception ex)
    {
        m_dsm_api_wrapper.reset();
        throw ex;
    }

    m_bus_state = m_dsm_api_wrapper->getBusState();

    m_dsm_api_wrapper->addListener(this);
}

bus_state_t dS_Assistant::getBusState() const
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    return m_bus_state;
}

boost::shared_ptr<dSMApiWrapper> dS_Assistant::dS485()
{
    if (m_dsm_api_wrapper == NULL)
    {
        throw _Exception("no active connection to dS485 daemon available");
    }
    return m_dsm_api_wrapper;
}

void dS_Assistant::addTask(boost::shared_ptr<dSTask> task)
{
    lockTaskList();
    m_task_list.push_front(task);
    pthread_cond_signal(&m_wakeup_task_cond);
    unlockTaskList();
}

long long int dS_Assistant::addLongTask(boost::shared_ptr<dSLongTask> task)
{
    lockLongTaskList();
#ifndef ADVANCED_TASK_MANAGEMENT
    if (m_current_task_info != NULL)
    {
        m_current_task_info->invalidate();
    }
#endif
    long long int task_id = Database::getInstance()->saveTask(task);
    task->setTaskID(task_id);

    m_long_task_list.push_front(task);
    pthread_cond_signal(&m_wakeup_long_task_cond);
    unlockLongTaskList();
    return task_id;
}

///****************************************************************************

void dS_Assistant::busStateCallback(bus_state_t state)
{
    m_bus_state = state;
    FILE_LOG(logINFO) << "dS485 Bus State: " << state;
}


void dS_Assistant::busChangeCallback(dsid_t *id, int flag)
{
    try
    {
        if (!DsmApiIsdSM(*id))
        {
            return;
        }

        FILE_LOG(logDEBUG1) << "dS485 Bus Change: " << util::str(*id) <<
                               " - " << flag;

        boost::shared_ptr<const dSM> dsmo = Storage::getInstance()->getDSM(*id);
        boost::shared_ptr<dSM> dsm(dsmo == NULL ? new dSM(*id) : dsmo->clone());

        if (flag)
        {
            DSID_LOG(logINFO, util::str(*id)) << "left";
            dsm->setPresent(0);
        }
        else
        {
            DSID_LOG(logINFO, util::str(*id)) << "joined";
            dsm->setPresent(1);
            boost::shared_ptr<GetDSMInfoTask> task(
                    new GetDSMInfoTask(dsm->getDSID()));
            addTask(boost::dynamic_pointer_cast<dSTask>(task));
        }
        Storage::getInstance()->addDSM(dsm);
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback: " <<
                              ex.getMessage();
        //FILE_LOG(logERROR) << ex.getStackTrace();
    }
    catch (std::exception e)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback: " << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback";
    }
}

///****************************************************************************

void dS_Assistant::deviceAddedCallback(uint8_t error_code,
                                       dsid_t src_dsm_dsid,
                                       dsid_t dest_dsm_dsid,
                                       uint16_t device_id,
                                       uint16_t zone_id,
                                       uint16_t vendor_id,
                                       uint32_t serial_number)
{
    dsid_t dsid;
    if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &dsid) != 0)
    {
        FILE_LOG(logDEBUG2) << "Ignoring unidentified device on dSM " <<
                               util::str(src_dsm_dsid);
        return;
    }

    boost::shared_ptr<GetDeviceInfoTask> task(
                            new GetDeviceInfoTask(src_dsm_dsid, device_id));
    addTask(boost::dynamic_pointer_cast<dSTask>(task));

    DSID_LOG(logINFO, util::str(dsid)) << "active";
}

void dS_Assistant::deviceRemovedCallback(uint8_t error_code,
                                         dsid_t src_dsm_dsid,
                                         dsid_t dest_dsm_dsid,
                                         uint16_t device_id,
                                         uint16_t zone_id,
                                         uint16_t vendor_id,
                                         uint32_t serial_number)
{
    dsid_t dsid;
    if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &dsid) != 0)
    {
        FILE_LOG(logDEBUG2) << "Ignoring unidentified device on dSM " <<
                               util::str(src_dsm_dsid);
        return;
    }

    boost::shared_ptr<const Device> dsdo =
                        Storage::getInstance()->getDevice(dsid, src_dsm_dsid);
    if (dsdo)
    {
        boost::shared_ptr<Device> dsd(dsdo->clone());
        dsd->setState(0);
        Storage::getInstance()->addDevice(dsd);
    }
    DSID_LOG(logINFO, util::str(dsid)) << "inactive";
}

void dS_Assistant::deviceStatisticsCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t offset, uint16_t value)
{
    boost::shared_ptr<const Device> dsdo =
                Storage::getInstance()->getDevice(device_id, src_dsm_dsid);
    if (dsdo)
    {
        DSID_LOG(logINFO, util::str(dsdo->getDSID())) <<
            "statistic offset:" << util::hex(offset) <<
            ", value:" << util::hex(value);
    }
}

void dS_Assistant::deviceActionCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
        uint8_t click_type)
{

}

void dS_Assistant::deviceLocalActionCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t state)
{

}

void dS_Assistant::deviceConfigCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t value)
{
    boost::shared_ptr<const Device> dsdo =
                    Storage::getInstance()->getDevice(device_id, src_dsm_dsid);
    if (dsdo)
    {
        DSID_LOG(logINFO, util::str(dsdo->getDSID())) <<
            "device config class:" << util::hex(config_class) <<
            ", index:" << util::hex(config_index) <<
            ", value:" << util::hex(value);
    }
}

void dS_Assistant::deviceTestTransmissionQualityCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t downstream_quality, uint16_t upstream_quality)
{
    boost::shared_ptr<const Device> dsdo = Storage::getInstance()->getDevice(device_id, src_dsm_dsid);
    if (dsdo)
    {
        DSID_LOG(logINFO, util::str(dsdo->getDSID())) <<
            "test quality down:" << util::hex(downstream_quality) <<
            ", up:" << util::hex(upstream_quality);
    }
}

///****************************************************************************

void dS_Assistant::commUpstreamShortCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
        uint8_t key_type, uint8_t click_type,
        uint8_t quality, uint8_t flags)
{
    DSID_LOG(logINFO, util::str(src_dsm_dsid)) <<
        "[Q:" << std::setprecision (2) <<
        (quality * 100.0/62 + .5) << ", P:" << ((flags & 1) > 0 ? 'N' : 'P') <<
        "] " <<
        dsutil::rkDecodeShort(src_dsm_dsid, device_id, is_resend, is_sensor,
                              key_type, click_type);
}

void dS_Assistant::commUpstreamLongCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier, uint32_t data,
        uint8_t quality, uint8_t flags)
{
    DSID_LOG(logINFO, util::str(src_dsm_dsid)) <<
        "[Q:" << std::setprecision (2) <<
        (quality * 100.0/62 + .5) << ", P:" <<
        ((flags & 1) > 0 ? 'N' : 'P') << "] " <<
        dsutil::rkDecodeLong(src_dsm_dsid, device_id, mainqualifier,
                             subqualifier, data);
}

void dS_Assistant::commDownstreamShortCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    DSID_LOG(logINFO, util::str(src_dsm_dsid)) <<
            dsutil::hkDecodeShort(command);
}

void dS_Assistant::commDownstreamLong0Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    DSID_LOG(logINFO, util::str(src_dsm_dsid)) <<
            dsutil::hkDecodeLong0(command);
}

void dS_Assistant::commDownstreamLong8Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
        uint8_t data)
{
    DSID_LOG(logINFO, util::str(src_dsm_dsid)) <<
            dsutil::hkDecodeLong8(command, data);
}

void dS_Assistant::commDownstreamLong16Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
        uint16_t data)
{
    DSID_LOG(logINFO, util::str(src_dsm_dsid)) <<
            dsutil::hkDecodeLong16(command, data);
}

///****************************************************************************

void dS_Assistant::lockTaskList()
{
    int ret = pthread_mutex_lock(&m_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to lock task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::unlockTaskList()
{
    int ret = pthread_mutex_unlock(&m_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to unlock task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::lockLongTaskList() const
{
    int ret = pthread_mutex_lock(&m_long_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to lock long task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::unlockLongTaskList() const
{
    int ret = pthread_mutex_unlock(&m_long_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to unlock long task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::run()
{
    int ret;

    if (m_running)
    {
        throw _Exception("Task threads are already running");
    }

    connect();

    m_running = true;

    try
    {
        clearPendingTasks();
    }
    catch (Exception ex)
    {
        m_running = false;
        throw ex;
    }

    ret = pthread_create(&m_task_thread, NULL, dS_Assistant::staticThreadProc,
                         this);
    if (ret != 0)
    {
        disconnect();
        m_running = false;
        throw _Exception("Could not start task thread");
    }

    ret = pthread_create(&m_long_task_thread, NULL,
                         dS_Assistant::staticLongTaskThreadProc, this);
    if (ret != 0)
    {
        pthread_join(m_task_thread, NULL);
        disconnect();
        m_running = false;
        throw _Exception("Could not start long task thread");
    }

    Database::getInstance()->updateTaskStatus(Running, Cancelled);
#ifdef ADVANCED_TASK_MANAGEMENT
    scan(false); // synchronous scan

    boost::shared_ptr<std::list<boost::shared_ptr<dSLongTask> > > tasks =
        Database::getInstance()->getTasksByStatus(Queued);

    if (tasks->empty())
    {
        return;
    }

    lockLongTaskList();
    while (!tasks->empty())
    {
        boost::shared_ptr<dSLongTask> task = tasks->front();
        tasks->pop_front();
        m_long_task_list.push_front(task);

    }
    pthread_cond_signal(&m_wakeup_long_task_cond);
    unlockLongTaskList();
#else
    Database::getInstance()->removeTasksByStatus(Queued);
    scan(true); // asynchronous scan
#endif
}

void dS_Assistant::taskThread()
{
    boost::shared_ptr<dSTask> task;

    lockTaskList();
    while (!m_shutdown_flag)
    {
        if (m_task_list.empty())
        {
            pthread_cond_wait(&m_wakeup_task_cond, &m_task_list_mutex);
            continue;
        }

        task = m_task_list.back();
        m_task_list.pop_back();

        unlockTaskList();

        try
        {
            if (task->isValid())
            {
                task->run(this);
            }
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) << "Task thread caught exception: " <<
                ex.getMessage();
            //FILE_LOG(logERROR) << ex.printStackTrace();
        }
        catch (std::exception e)
        {
            FILE_LOG(logERROR) << "Task thread caught exception: " << e.what();
        }
        catch (...)
        {
            FILE_LOG(logERROR) << "Task thread caught unknown exception";
        }

        task.reset();

        if (!m_shutdown_flag)
        {
            lockTaskList();
        }
    }
    unlockTaskList();
}

void *dS_Assistant::staticThreadProc(void *arg)
{
    dS_Assistant *instance = (dS_Assistant *)arg;
    instance->taskThread();
    pthread_exit(NULL);
    return NULL;
}

void dS_Assistant::longTaskThread()
{
    boost::shared_ptr<dSLongTask> task;

    lockLongTaskList();
    while (!m_shutdown_flag)
    {
        if (m_long_task_list.empty())
        {
            pthread_cond_wait(&m_wakeup_long_task_cond,
                              &m_long_task_list_mutex);
            continue;
        }

        task = m_long_task_list.back();
        m_long_task_list.pop_back();

        m_current_task_info = boost::dynamic_pointer_cast<dSLongTaskInfo>(task);
        unlockLongTaskList();

        try
        {
            if (task->isValid())
            {
                Database::getInstance()->updateTaskStatus(task->getTaskID(),
                                                          Running);
                task->run(this);
                if (task->isValid())
                {
                    Database::getInstance()->updateTaskStatus(task->getTaskID(),
                                                              Finished);
                }
                else
                {
                    Database::getInstance()->updateTaskStatus(task->getTaskID(),
                                                              Cancelled);
                }
            }
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) << "Long task thread caught exception: " <<
                ex.getMessage();
            //FILE_LOG(logERROR) << ex.printStackTrace();
        }
        catch (std::exception e)
        {
            FILE_LOG(logERROR) << "Long task thread caught exception: " <<
                e.what();
        }
        catch (...)
        {
            FILE_LOG(logERROR) << "Long task thread caught unknown exception";
        }

        task.reset();

        if (!m_shutdown_flag)
        {
            lockLongTaskList();
            m_current_task_info.reset();
        }
    }
    unlockLongTaskList();
}

void *dS_Assistant::staticLongTaskThreadProc(void *arg)
{
    dS_Assistant *instance = (dS_Assistant *)arg;
    instance->longTaskThread();
    pthread_exit(NULL);
    return NULL;
}

void dS_Assistant::scan(bool async)
{
    boost::shared_ptr<GetBusMembersTask> task(new GetBusMembersTask(async));
    if (async)
    {
        addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
    else
    {
        task->run(this);
    }
}

void dS_Assistant::clearPendingTasks()
{
    lockTaskList();
    while (!m_task_list.empty())
    {
        boost::shared_ptr<dSTask> task = m_task_list.front();
        task.reset();
        m_task_list.pop_front();
    }
    unlockTaskList();

    lockLongTaskList();
    while (!m_long_task_list.empty())
    {
        boost::shared_ptr<dSLongTask> task = m_long_task_list.front();
        task.reset();
        m_long_task_list.pop_front();
    }
    unlockLongTaskList();
}

boost::shared_ptr<const dSLongTaskInfo> dS_Assistant::getCurrentTaskInfo() const
{
    lockLongTaskList();
    boost::shared_ptr<const dSLongTaskInfo> ret = m_current_task_info;
    unlockLongTaskList();
    return ret;
}

boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
                                        dS_Assistant::getTaskList() const
{
    lockLongTaskList();
    boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
        taskList(new std::list<boost::shared_ptr<const dSLongTaskInfo> >);

    if (m_current_task_info != NULL)
    {
        taskList->push_back(m_current_task_info);
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_long_task_list.begin(); it != m_long_task_list.end(); it++)
    {
        boost::shared_ptr<const dSLongTaskInfo> task =
                            boost::dynamic_pointer_cast<dSLongTaskInfo>(*it);
        taskList->push_back(task);
    }

    unlockLongTaskList();

    return taskList;
}

void dS_Assistant::cancelTask(long long int task_id)
{
    bool found = false;

    lockLongTaskList();
    if ((m_current_task_info != NULL) &&
       (m_current_task_info->getTaskID() == task_id))
    {
        found = true;
        m_current_task_info->invalidate();
    }
    else
    {
        for (std::list<boost::shared_ptr<dSLongTask> >::iterator
            it = m_long_task_list.begin(); it != m_long_task_list.end(); it++)
        {
            boost::shared_ptr<dSLongTask> task = *it;
            if (task->getTaskID() == task_id)
            {
                found = true;
                task->invalidate();
                Database::getInstance()->updateTaskStatus(task_id, Cancelled);
                m_long_task_list.erase(it);
                break;
            }
        }
    }
    unlockLongTaskList();
    if (!found)
    {
        throw _Exception("task with id " + util::str(task_id) + " not found");
    }
}

void dS_Assistant::disconnect()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    if (m_dsm_api_wrapper)
    {
        try
        {
            clearPendingTasks();
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) << "Failed to clear pending tasks: " <<
                ex.getMessage();
        }
        m_dsm_api_wrapper->removeListener(this);
        m_dsm_api_wrapper.reset();
    }
}

void dS_Assistant::shutdown()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    m_shutdown_flag = true;

    clearPendingTasks();
    lockTaskList();
    pthread_cond_signal(&m_wakeup_task_cond);
    unlockTaskList();
    if (m_task_thread)
    {
        pthread_join(m_task_thread, NULL);
        m_task_thread = 0;
    }

    lockLongTaskList();
    if (m_current_task_info != NULL)
    {
        m_current_task_info->invalidate();
    }
    pthread_cond_signal(&m_wakeup_long_task_cond);
    unlockLongTaskList();
    if (m_long_task_thread)
    {
        pthread_join(m_long_task_thread, NULL);
        m_long_task_thread = 0;
    }

    lock.unlock();
    disconnect();
}

const dsid_t dS_Assistant::getDSID() const
{
    return m_own_dsid;
}

long long int dS_Assistant::getInvalidTaskID() const
{
    return INVALID_TASK_ID;
}
dS_Assistant::~dS_Assistant()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    m_instance = NULL;
    lock.unlock();
    try
    {
        shutdown();
        disconnect();
        Database::getInstance()->updateTaskStatus(Running, Cancelled);
#ifndef ADVANCED_TASK_MANAGEMENT
        Database::getInstance()->removeTasksByStatus(Queued);
#endif
    }
    catch (...)
    {
        FILE_LOG(logERROR) << "Exception caught when shutting down, ignoring";
    }

    pthread_mutex_destroy(&m_task_list_mutex);
    pthread_mutex_destroy(&m_long_task_list_mutex);
    pthread_cond_destroy(&m_wakeup_task_cond);
    pthread_cond_destroy(&m_wakeup_long_task_cond);
}

dSTaskInfo::dSTaskInfo() : m_name("Generic task"), m_description(""),
                           m_valid(false),
                           m_cancellable(true) {}

void dSTaskInfo::setDescription(std::string description)
{
    m_description = description;
}

const std::string dSTaskInfo::getName() const
{
    return m_name;
}

const std::string dSTaskInfo::getDescription() const
{
    return m_description;
}

bool dSTaskInfo::isValid() const
{
    return m_valid;
}

void dSTaskInfo::invalidate()
{
    if (m_cancellable)
    {
        m_valid = false;
    }
}

dSTaskInfo::~dSTaskInfo() {}

bool dSTaskInfo::isCancellable() const
{
    return m_cancellable;
}

dSTask::dSTask() : dSTaskInfo() {}

dSLongTaskInfo::dSLongTaskInfo() : dSTaskInfo(), m_task_id(INVALID_TASK_ID),
                                    m_parent_task_id(INVALID_TASK_ID),
                                    m_progress(0),
                                    m_status_info("Queued"),
                                    m_task_status(Queued),
                                    m_started(0),
                                    m_finished(0),
                                    m_queued(0) {}

void dSLongTaskInfo::setTaskID(long long int task_id)
{
    m_task_id = task_id;
}

long long int dSLongTaskInfo::getTaskID() const
{
    return m_task_id;
}

void dSLongTaskInfo::setParentTaskID(long long int task_id)
{
    m_parent_task_id = task_id;
}

long long int dSLongTaskInfo::getParentTaskID() const
{
    return m_parent_task_id;
}

unsigned char dSLongTaskInfo::getProgress() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_progress;
}

std::string dSLongTaskInfo::getStatusInfo() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_status_info;
}

std::string dSLongTaskInfo::getParameters() const
{
    return m_parameters;
}

void dSLongTaskInfo::setTaskStatus(task_status status)
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    m_task_status = status;
    switch (status)
    {
        case Running:
            m_status_info = "läuft gerade";
            break;
        case Queued:
            m_status_info = "gequeued";
            break;
        case Cancelled:
            m_status_info = "abgebrochen";
            break;
        case Finished:
            m_status_info = "beendet";
            break;
        case UnknownStatus:
        default:
            m_status_info = "unbekannt";
            break;
    }
}

task_status dSLongTaskInfo::getTaskStatus() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_task_status;
}

std::string dSLongTaskInfo::getTaskStatusText() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    if (m_task_status == Running)
    {
        return "running";
    }
    else if (m_task_status == Cancelled)
    {
        return "cancelled";
    }
    else if (m_task_status == Finished)
    {
        return "finished";
    }
    else if (m_task_status == Queued)
    {
        return "queued";
    }
    else
    {
        return "unknown";
    }
}

time_t dSLongTaskInfo::getStartedTime() const
{
    return m_started;
}

void dSLongTaskInfo::setStartedTime(time_t t)
{
    m_started = t;
}

time_t dSLongTaskInfo::getFinishedTime() const
{
    return m_finished;
}

void dSLongTaskInfo::setFinishedTime(time_t t)
{
    m_finished = t;
}

time_t dSLongTaskInfo::getQueuedTime() const
{
    return m_queued;
}

void dSLongTaskInfo::setQueuedTime(time_t t)
{
    m_queued = t;
}

dSLongTask::dSLongTask() : dSLongTaskInfo() {}

GetBusMembersTask::GetBusMembersTask(bool async) : dSTask(), m_async(async)
{
    m_name = "Get bus members";
    m_description = "Scan the dS485 bus in order to retrieve information "
                    "about all available devices";
    m_valid = true;
}


void GetBusMembersTask::run(dS_Assistant *dsa)
{
    dsid_t device_list[DSM_API_MAX_DEVICES];

    if (!(dsa && m_valid)) {
        return;
    }

    int devices = dsa->dS485()->getBusMembers(device_list, DSM_API_MAX_DEVICES);
    for (int i = 0; i < devices; ++i)
    {
        dsid_t own_dsid = dsa->getDSID();
        if (IsEqualId(device_list[i], own_dsid))
        {
            continue;
        }

        if (DsmApiIsdSM(device_list[i]))
        {
            boost::shared_ptr<dSM> dsm(new dSM(device_list[i]));
            Storage::getInstance()->addDSM(dsm);
            boost::shared_ptr<GetDSMInfoTask> task(
                                  new GetDSMInfoTask(dsm->getDSID(), m_async));
            if (m_async)
            {
                dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
            }
            else
            {
                task->run(dsa);
            }
        }
        else
        {
            boost::shared_ptr<HWBase> h(new HWBase(device_list[i]));
            Storage::getInstance()->add(h);
        }
    }
}

GetDSMInfoTask::GetDSMInfoTask(dsid_t dsid, bool async) : dSTask(),
                                                          m_dsid(dsid),
                                                          m_async(async)
{
    m_name = "Request dSM Information";
    m_description = "Retrieve information for dSM " + util::str(dsid);
    m_valid = true;
}

void GetDSMInfoTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint32_t hwv;
    uint32_t armv;
    uint32_t dspv;
    uint16_t apiv;
    dsid_t dsid;
    uint8_t name[21];
    uint32_t cHash, cVersion;

    boost::shared_ptr<const dSM> dsmo = Storage::getInstance()->getDSM(m_dsid);
    boost::shared_ptr<dSM> dsm(dsmo == NULL ? new dSM(m_dsid) : dsmo->clone());

    dsa->dS485()->dSMInfo(m_dsid, &hwv, &armv, &dspv, &apiv, &dsid, name);
    dsa->dS485()->getConfigurationId(m_dsid, &cHash, &cVersion);

    dsm->setPresent(1);
    dsm->setHWVersion(hwv);
    dsm->setARMSwVersion(armv);
    dsm->setDSPSwVersion(dspv);
    dsm->setAPIVersion(apiv);
    dsm->setName(std::string((const char*)name));

    uint64_t configId = (uint64_t) cHash << 32 | cVersion;
    if (configId != dsm->getConfigurationId())
    {
        boost::shared_ptr<ScanDSMTask> task(new ScanDSMTask(dsid, m_async));
        dsm->setConfigurationId(configId);
        if (m_async)
        {
            dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
        }
        else
        {
            task->run(dsa);
        }
    }

    Storage::getInstance()->addDSM(dsm);

    DSID_LOG(logINFO, m_dsid) << "[Task DSMInfo] ARM:" << std::hex <<
        armv << ", DSP:" << dspv << ", HW:" << hwv << ", CFG:" << configId;

    // do one-time initialization stuff each time a dsm joins the bus
    // TODO: move this to a common dsm method
    // TODO: configure 24h ping mode every night at 3 a.m., through external cron job?
    dsa->dS485()->setCommunicationLogging(m_dsid, 0xff);
}

ScanDSMTask::ScanDSMTask(dsid_t dsid, bool async) : dSTask(), m_dsid(dsid),
                                                    m_async(async)
{
    m_name = "Scanning dSM for devices";
    m_description = "Retrieve information about devices attached to the "
                    "dS-Meter";
    m_valid = true;
}

void ScanDSMTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    boost::shared_ptr<const dSM> dsmo = Storage::getInstance()->getDSM(m_dsid);
    if (dsmo == NULL)
    {
        throw _Exception("dSM " + util::str(m_dsid) + " not found");
    }
    boost::shared_ptr<dSM> dsm(dsmo->clone());

    uint8_t value;
    dsa->dS485()->getApartmentId(m_dsid, &value);
    dsm->setApartmentId(value);

    Storage::getInstance()->addDSM(dsm);
    uint8_t count;
    dsa->dS485()->getZoneCount(m_dsid, &count);
    for (uint8_t i = 0; i < count; i++)
    {
        uint16_t zone_id;
        dsa->dS485()->getZoneInfo(m_dsid, i, &zone_id, NULL, NULL, NULL);
        boost::shared_ptr<GetActiveDevicesInZoneTask> task(
                            new GetActiveDevicesInZoneTask(m_dsid, zone_id));
        if (m_async)
        {
            dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
        }
        else
        {
            task->run(dsa);
        }
    }


}

GetActiveDevicesInZoneTask::GetActiveDevicesInZoneTask(dsid_t dsid,
                                                       uint16_t zone_id) :
                                                       dSTask(), m_dsid(dsid),
                                                       m_zone_id(zone_id)
{
    m_name = "Getting active devices from zone";
    m_description = "Retrieve information about active devices in a given zone"
                    " on a dS-Meter";
    m_valid = true;
}

void GetActiveDevicesInZoneTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t count;

    uint16_t device_id;
    uint16_t vendor_id;
    uint16_t product_id;
    uint16_t function_id;
    uint16_t version;
    uint16_t zone_id;
    uint8_t state;
    uint8_t my_click;
    uint8_t locked;
    uint8_t output_mode;
    uint8_t groups[8];
    uint8_t name[21];
    uint8_t buttonGroups[16];
    uint8_t button_zone_membership[2];
    uint32_t serial_number;

    dsa->dS485()->getActiveDevicesCountInZone(m_dsid, m_zone_id, &count);
    for (uint16_t i = 0; i < count; i++)
    {
        dsa->dS485()->getActiveDeviceInfoInZone(m_dsid, m_zone_id, i,
                                       &device_id,
                                       &vendor_id, &product_id, &function_id,
                                       &version, &zone_id, &state, &my_click,
                                       &locked, &output_mode, groups,
                                       name, buttonGroups,
                                       button_zone_membership, &serial_number);

        dsid_t device_dsid;
        if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &device_dsid) != 0)
        {
            FILE_LOG(logERROR) <<
                "[Task GetActiveDevices] Could not resolve dsid: "
                "serial " << util::hex(device_id) <<
                "/vendor " << util::hex(vendor_id);
            continue;
        }

        boost::shared_ptr<Device> dev(new Device(device_dsid, m_dsid));
        dev->setId(device_id);
        dev->setProductId(product_id);
        dev->setFunctionId(function_id);
        dev->setVersion(version);
        dev->setZoneId(zone_id);
        dev->setState(state);
        dev->setMyClick(my_click);
        dev->setLocked(locked);
        dev->setOutputMode(output_mode);
        dev->setName((const char*)name);
        dev->setSerialNumber(serial_number);
        Storage::getInstance()->addDevice(dev);

        FILE_LOG(logINFO) << "Added device with dSID " <<
            util::str(dev->getDSID()) << " on dSM " <<
            util::str(dev->getDSMdSID());
    }
}

GetDSMZonesTask::GetDSMZonesTask(dsid_t dsid) : dSTask(), m_dsid(dsid)
{
    m_name = "Getting number of zones on dSM";
    m_description = "Retrieve number of zones for dSM " + util::str(dsid);
    m_valid = true;
}

void GetDSMZonesTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint8_t count;
    dsa->dS485()->getZoneCount(m_dsid, &count);
    for (uint8_t i = 0; i < count; i++)
    {
        boost::shared_ptr<GetDSMZoneInfoTask> task(
                                            new GetDSMZoneInfoTask(m_dsid, i));
        dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
}

GetDSMZoneInfoTask::GetDSMZoneInfoTask(dsid_t dsid, uint8_t zone_index) :
                                       dSTask(), m_dsid(dsid),
                                       m_zone_index(zone_index)
{
    m_name = "Getting zone information";
    m_description = "Retrieve information for zone with index " +
                     util::str(zone_index) + " on dSM " + util::str(dsid);
    m_valid = true;
}

void GetDSMZoneInfoTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t zone_id;
    uint8_t vzone_id;
    uint8_t num_grp;
    uint8_t name[21];

    dsa->dS485()->getZoneInfo(m_dsid, m_zone_index, &zone_id, &vzone_id,
                              &num_grp, name);
    printf("Retrieved zone info for dsm: %s, zoneid: %u vzid: %u groups: %u "
           "name %s\n", util::str(m_dsid).c_str(), zone_id, vzone_id,
           num_grp, name);
}

GetDeviceInfoTask::GetDeviceInfoTask(dsid_t dsm_dsid, uint16_t device_id) :
                                     dSTask(), m_dsid(dsm_dsid),
                                     m_device_id(device_id)
{
    m_name = "Getting device information";
    m_description = "Retrieve device information for device on dSM " +
                   util::str(dsm_dsid);
    m_valid = true;
}

void GetDeviceInfoTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t device_id;
    uint16_t vendor_id;
    uint16_t product_id;
    uint16_t function_id;
    uint16_t version;
    uint16_t zone_id;
    uint8_t state;
    uint8_t my_click;
    uint8_t locked;
    uint8_t output_mode;
    uint8_t groups[8];
    uint8_t name[21];
    uint8_t buttonGroups[16];
    uint8_t button_zone_membership[2];
    uint32_t serial_number;

    dsa->dS485()->getDeviceInfo(m_dsid, m_device_id,
                       &device_id, &vendor_id, &product_id, &function_id,
                       &version, &zone_id, &state, &my_click, &locked,
                       &output_mode, groups, name, buttonGroups,
                       button_zone_membership, &serial_number);

    dsid_t device_dsid;
    if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &device_dsid) != 0)
    {
        printf("Could not resolve dsid: vendor id is not known, ignoring "
               "device");
        return;
    }

    boost::shared_ptr<Device> dev(new Device(device_dsid, m_dsid));
    dev->setId(device_id);
    dev->setProductId(product_id);
    dev->setFunctionId(function_id);
    dev->setVersion(version);
    dev->setZoneId(zone_id);
    dev->setState(state);
    dev->setMyClick(my_click);
    dev->setLocked(locked);
    dev->setOutputMode(output_mode);
    dev->setName((const char*)name);
    Storage::getInstance()->addDevice(dev);

    FILE_LOG(logINFO) << "Added device with dSID " << util::str(dev->getDSID()) <<  " on dSM " << util::str(dev->getDSMdSID());
}

TestTranmissionQualityTask::TestTranmissionQualityTask(int checksPerDevice,
                                dsid_t dsm_dsid) : dSLongTask(),
                                m_checksPerDevice(checksPerDevice),
                                m_dsm_dsid(dsm_dsid)
{
    m_name = "Test der Übertragungsqualität der Klemmen";
    m_description = "Testen der Übertragungsqualität für ";
    m_parameters = "dsm=" + util::str(dsm_dsid) + '&' +
                   "checksPerDevice=" + util::str(checksPerDevice);

    if (IsNullId(dsm_dsid))
    {
        m_description += "alle Klemmen";
    }
    else
    {
        m_description += "alle Klemmen am dSM " + util::str(dsm_dsid);
    }

    m_description += ", Tests pro Klemme: " + util::str(checksPerDevice);

    m_valid = true;
}

task_type TestTranmissionQualityTask::getTaskType() const
{
    return TestTransmissionQuality;
}

void TestTranmissionQualityTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    boost::shared_ptr<std::list<boost::shared_ptr<const Device> > > devs;
    if (IsNullId(m_dsm_dsid))
    {
        devs = Storage::getInstance()->getDevices();
    }
    else
    {
        devs = Storage::getInstance()->getDevices(m_dsm_dsid);
    }

    size_t total = devs->size() * m_checksPerDevice;
    size_t current = 0;
    int check_loop = 0;

    m_status_mutex.lock();
    m_progress = (unsigned char)(current * 100 / total);
    m_status_mutex.unlock();

        std::list<boost::shared_ptr<const Device> >::const_iterator it =
                                                                devs->begin();

    while (m_valid && (current < total))
    {
        boost::shared_ptr<const Device> device = *it;

        current++;

        m_status_mutex.lock();
        m_progress = (unsigned char)(current * 100 / total);
        m_status_info = "Requesting transmission quality from device " +
                        util::str(device->getDSID());
        if (m_checksPerDevice > 1)
        {
            m_status_info += ", test " + util::str(check_loop + 1);
        }
        m_status_mutex.unlock();

        uint8_t down = TRANSMISSION_QUALITY_DOWNSTREAM_MAX;
        uint16_t up = 0;
        int err = 0;
        try
        {
            dsa->dS485()->testDeviceTransmissionQuality(device->getDSMdSID(),
                                               device->getId(), &down, &up);

            FILE_LOG(logINFO) << "Transmission quality for device " <<
                util::str(device->getDSID()) <<
                ", up " << util::str(up) << ", down " << util::str(down);
        }
        catch (Exception ex)
        {
            m_status_mutex.lock();
            m_status_info = "Failed to get transmission quality for "
                            "device " + util::str(device->getDSID()) +
                            ": " + ex.getMessage();
            printf("%s\n", m_status_info.c_str());
            err = ex.getErrorCode();
            m_status_mutex.unlock();
        }

        Database::getInstance()->saveTransmissionQuality(m_task_id,
                                util::str(device->getDSID()), up, down, err);

        it++;
        if (it == devs->end())
        {
            check_loop++;
            if (check_loop < m_checksPerDevice)
            {
                it = devs->begin();
            }
            else
            {
                break;
            }
        }
    }

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = "Cancelled";
    }
    else
    {
        m_task_status = Finished;
        m_status_info = "Finished";
    }
    m_status_mutex.unlock();
}

} // namespace
