/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#include <digitalSTROM/ds485-client-interface.h>

#include "dsmapi-wrapper.h"
#include "exceptions.h"
#include "traffic-decode.h"
#include "util.h"

#define DEFAULT_SYNC_CALL_TIMEOUT   5 // seconds

/// \file dsmapi-wrapper.h
namespace dSA
{

dSMApiWrapper::dSMApiWrapper(std::string ds485_daemon_url)
{
    int ret;

    boost::mutex::scoped_lock lock(m_libdsm_mutex);

    m_dsm_api_handle = DsmApiInitialize();
    if (m_dsm_api_handle == NULL)
    {
        throw _Exception("could not initialize dSM API library");
    }

    ret = DsmApiOpen(m_dsm_api_handle, ds485_daemon_url.c_str(), 0);
    if (ret < 0)
    {
        throw _Exception("could not open connection to " + ds485_daemon_url +
                         " (" + util::str(ret) + ")");
    }

    // register callbacks
    DsmApiRegisterBusStateCallback(m_dsm_api_handle,
                                   dSMApiWrapper::busStateCallback, this);

    DsmApiRegisterBusChangeCallback(m_dsm_api_handle,
                                    dSMApiWrapper::busChangeCallback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_ACCESSIBILITY,
                           EVENT_DEVICE_ACCESSIBILITY_ON,
                           (void *)dSMApiWrapper::deviceAddedCallback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_ACCESSIBILITY,
                           EVENT_DEVICE_ACCESSIBILITY_OFF,
                           (void *)dSMApiWrapper::deviceRemovedCallback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_STATISTICS,
                           0,
                           (void *)dSMApiWrapper::deviceStatisticsCallback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_ACTION,
                           0,
                           (void *)dSMApiWrapper::deviceActionCallback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_LOCAL_ACTION,
                           0,
                           (void *)dSMApiWrapper::deviceLocalActionCallback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_CONFIG,
                           0,
                           (void *)dSMApiWrapper::deviceConfigCallback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_TEST_TRANSMISSION_QUALITY,
                           0,
                           (void *)dSMApiWrapper::deviceTestTransmissionQualityCallback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_UPSTREAM_SHORT,
                           (void *)dSMApiWrapper::commUpstreamShortCallback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_UPSTREAM_LONG,
                           (void *)dSMApiWrapper::commUpstreamLongCallback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_DOWNSTREAM_SHORT,
                           (void *)dSMApiWrapper::commDownstreamShortCallback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_DOWNSTREAM_LONG_NO_PARAMETER,
                           (void *)dSMApiWrapper::commDownstreamLong0Callback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_DOWNSTREAM_LONG_8BIT_PARAMETER,
                           (void *)dSMApiWrapper::commDownstreamLong8Callback, this);

    DsmApiRegisterCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_DOWNSTREAM_LONG_16BIT_PARAMETER,
                           (void *)dSMApiWrapper::commDownstreamLong16Callback, this);
}

void dSMApiWrapper::addListener(dSMApiListener* listener)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    m_callback_listeners.push_back(listener);
}

void dSMApiWrapper::removeListener(dSMApiListener* listener)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == listener)
        {
            i = m_callback_listeners.erase(i);
            break;
        }

        i++;
    }
}

void dSMApiWrapper::busStateCallback(void* arg, bus_state_t state)
{
    static_cast<dSMApiWrapper *>(arg)->notifyBusState(state);
}

void dSMApiWrapper::notifyBusState(bus_state_t state)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->busStateCallback(state);
            i++;
        }
    }
}

void dSMApiWrapper::busChangeCallback(void* arg, dsid_t *id, int flag)
{
    static_cast<dSMApiWrapper *>(arg)->notifyBusChange(id, flag);
}

void dSMApiWrapper::notifyBusChange(dsid_t *id, int flag)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->busChangeCallback(id, flag);
            i++;
        }
    }
}

void dSMApiWrapper::deviceAddedCallback(uint8_t error_code, void *arg,
                                       dsid_t src_dsm_dsid,
                                       dsid_t dest_dsm_dsid,
                                       uint16_t device_id,
                                       uint16_t zone_id,
                                       uint16_t vendor_id,
                                       uint32_t serial_number)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceAdded(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id, zone_id, vendor_id,
            serial_number);
}

void dSMApiWrapper::notifyDeviceAdded(uint8_t error_code, dsid_t src_dsm_dsid,
                                     dsid_t dest_dsm_dsid, uint16_t device_id,
                                     uint16_t zone_id, uint16_t vendor_id,
                                     uint32_t serial_number)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener *>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->deviceAddedCallback(error_code, src_dsm_dsid,
                                      dest_dsm_dsid, device_id,
                                      zone_id, vendor_id, serial_number);
            i++;
        }
    }
}

void dSMApiWrapper::deviceRemovedCallback(uint8_t error_code, void *arg,
                                         dsid_t src_dsm_dsid,
                                         dsid_t dest_dsm_dsid,
                                         uint16_t device_id,
                                         uint16_t zone_id,
                                         uint16_t vendor_id,
                                         uint32_t serial_number)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceRemoved(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id, zone_id, vendor_id,
            serial_number);
}

void dSMApiWrapper::notifyDeviceRemoved(uint8_t error_code, dsid_t src_dsm_dsid,
                                        dsid_t dest_dsm_dsid,
                                        uint16_t device_id, uint16_t zone_id,
                                        uint16_t vendor_id,
                                        uint32_t serial_number)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->deviceRemovedCallback(error_code, src_dsm_dsid,
                                        dest_dsm_dsid, device_id, zone_id,
                                        vendor_id, serial_number);
            i++;
        }
    }
}

void dSMApiWrapper::deviceStatisticsCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t offset, uint16_t value)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceStatistics(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id, offset, value);
}

void dSMApiWrapper::notifyDeviceStatistics(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t offset, uint16_t value)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->deviceStatisticsCallback(error_code,
                    src_dsm_dsid, dest_dsm_dsid, device_id, offset, value);
            i++;
        }
    }
}

void dSMApiWrapper::deviceActionCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t group_id, uint8_t button_nr, uint8_t click_type)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceActionCallback(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id,
            zone_id, group_id, button_nr, click_type);
}
void dSMApiWrapper::notifyDeviceActionCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t group_id, uint8_t button_nr, uint8_t click_type)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->deviceActionCallback(error_code,
                    src_dsm_dsid, dest_dsm_dsid, device_id,
                    zone_id, group_id, button_nr, click_type);
            i++;
        }
    }
}

void dSMApiWrapper::deviceLocalActionCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t state)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceLocalActionCallback(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id,
            zone_id, state);
}
void dSMApiWrapper::notifyDeviceLocalActionCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t state)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->deviceLocalActionCallback(error_code,
                    src_dsm_dsid, dest_dsm_dsid, device_id,
                    zone_id, state);
            i++;
        }
    }
}

void dSMApiWrapper::deviceConfigCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t value)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceConfigCallback(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id,
            config_class, config_index, value);
}

void dSMApiWrapper::notifyDeviceConfigCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t value)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->deviceConfigCallback(error_code,
                    src_dsm_dsid, dest_dsm_dsid, device_id,
                    config_class, config_index, value);
            i++;
        }
    }
}

void dSMApiWrapper::deviceTestTransmissionQualityCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t downstream_quality, uint16_t upstream_quality)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceTestTransmissionQualityCallback(error_code,
                src_dsm_dsid, dest_dsm_dsid, device_id,
                downstream_quality, upstream_quality);
}

void dSMApiWrapper::notifyDeviceTestTransmissionQualityCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t downstream_quality, uint16_t upstream_quality)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->deviceTestTransmissionQualityCallback(error_code,
                    src_dsm_dsid, dest_dsm_dsid, device_id,
                    downstream_quality, upstream_quality);
            i++;
        }
    }
}

void dSMApiWrapper::commUpstreamShortCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
        uint8_t key_type, uint8_t click_type,
        uint8_t quality, uint8_t flags)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommUpstreamShortCallback(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id,
            circuit_id, is_resend, is_sensor, key_type, click_type, quality, flags);
}

void dSMApiWrapper::notifyCommUpstreamShortCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
        uint8_t key_type, uint8_t click_type,
        uint8_t quality, uint8_t flags)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->commUpstreamShortCallback(error_code,
                    src_dsm_dsid, dest_dsm_dsid, device_id,
                    circuit_id, is_resend, is_sensor, key_type, click_type,
                    quality, flags);
            i++;
        }
    }
}

void dSMApiWrapper::commUpstreamLongCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier, uint32_t data,
        uint8_t quality, uint8_t flags)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommUpstreamLongCallback(error_code,
            src_dsm_dsid, dest_dsm_dsid, device_id,
            circuit_id, mainqualifier, subqualifier, data, quality, flags);
}

void dSMApiWrapper::notifyCommUpstreamLongCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier, uint32_t data,
        uint8_t quality, uint8_t flags)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->commUpstreamLongCallback(error_code,
                    src_dsm_dsid, dest_dsm_dsid, device_id,
                    circuit_id, mainqualifier, subqualifier, data,
                    quality, flags);
            i++;
        }
    }
}

void dSMApiWrapper::commDownstreamShortCallback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommDownstreamShortCallback(error_code,
            src_dsm_dsid, dest_dsm_dsid, command);
}

void dSMApiWrapper::notifyCommDownstreamShortCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->commDownstreamShortCallback(error_code,
                    src_dsm_dsid, dest_dsm_dsid, command);
            i++;
        }
    }
}

void dSMApiWrapper::commDownstreamLong0Callback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommDownstreamLong0Callback(error_code,
                src_dsm_dsid, dest_dsm_dsid, command);
}

void dSMApiWrapper::notifyCommDownstreamLong0Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->commDownstreamLong0Callback(error_code,
                    src_dsm_dsid, dest_dsm_dsid, command);
            i++;
        }
    }
}

void dSMApiWrapper::commDownstreamLong8Callback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command, uint8_t data)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommDownstreamLong8Callback(error_code,
                    src_dsm_dsid, dest_dsm_dsid, command, data);
}

void dSMApiWrapper::notifyCommDownstreamLong8Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command, uint8_t data)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->commDownstreamLong8Callback(error_code,
                    src_dsm_dsid, dest_dsm_dsid, command, data);
            i++;
        }
    }
}

void dSMApiWrapper::commDownstreamLong16Callback(uint8_t error_code, void *arg,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command, uint16_t data)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommDownstreamLong16Callback(error_code,
                        src_dsm_dsid, dest_dsm_dsid, command, data);
}

void dSMApiWrapper::notifyCommDownstreamLong16Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command, uint16_t data)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            (*i)->commDownstreamLong16Callback(error_code,
                    src_dsm_dsid, dest_dsm_dsid, command, data);
            i++;
        }
    }
}


void dSMApiWrapper::getOwnDSID(dsid_t* dsid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DsmApiGetOwnDSID(m_dsm_api_handle, dsid);
    checkRetcode(ret);
}

bus_state_t dSMApiWrapper::getBusState()
{
    bus_state_t state;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DsmApiGetBusState(m_dsm_api_handle, &state);
    checkRetcode(ret);
    return state;
}

void dSMApiWrapper::restartRegistration(dsid_t dsm_dsid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_restart_registration(m_dsm_api_handle, dsm_dsid);
    checkRetcode(ret);
}

void dSMApiWrapper::getDSMtemperature(dsid_t dsm_dsid, uint16_t *temperature)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMProperties_get_current_temperature(m_dsm_api_handle, dsm_dsid,
                                                    temperature);
    checkRetcode(ret);
}

int dSMApiWrapper::getBusMembers(dsid_t *device_list, int device_list_size)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DsmApiGetBusMembers(m_dsm_api_handle, device_list,
                                  DSM_API_MAX_DEVICES);
    if (ret < 0)
    {
        throw _ExceptionWithCode("could not get bus members", ret);
    }

    return ret;
}

void dSMApiWrapper::dSMInfo(dsid_t dsm_dsid, uint32_t *hw_version,
                           uint32_t *arm_sw_version, uint32_t *dsp_sw_version,
                           uint16_t *api_version, dsid_t *dsid, uint8_t *name)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    /// \todo make sure to always check the handle and connection state
    ret = ::dSMInfo(m_dsm_api_handle, dsm_dsid, hw_version, arm_sw_version,
                    dsp_sw_version, api_version, dsid, name);
    checkRetcode(ret);
}

void dSMApiWrapper::dSMState(dsid_t dsm_dsid, uint8_t *hardware,
                            uint8_t *communicationdsp, uint8_t *temperature,
                            uint8_t *configuration, uint8_t *ds485bus,
                            uint8_t *powerline)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ::dSMState(m_dsm_api_handle, dsm_dsid, hardware, communicationdsp,
                     temperature, configuration, ds485bus, powerline);
    checkRetcode(ret);
}

void dSMApiWrapper::getApartmentId(dsid_t dsm_dsid, uint8_t *id)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_get_apartment_id(m_dsm_api_handle, dsm_dsid, id);
    checkRetcode(ret);
}

void dSMApiWrapper::getConfigurationId(dsid_t dsm_dsid, uint32_t *hash, uint32_t *configcount)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMConfig_get_hash(m_dsm_api_handle, dsm_dsid, hash, configcount);
    checkRetcode(ret);
}

void dSMApiWrapper::getZoneCount(dsid_t dsm_dsid, uint8_t *count)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneCount(m_dsm_api_handle, dsm_dsid, count);
    checkRetcode(ret);
}

void dSMApiWrapper::getZoneInfo(dsid_t dsm_dsid, uint8_t zone_index,
                               uint16_t *zone_id, uint8_t *virtual_zone_id,
                               uint8_t *number_of_groups, uint8_t *name)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneInfo_by_index(m_dsm_api_handle, dsm_dsid, zone_index, zone_id,
                            virtual_zone_id, number_of_groups, name);
    checkRetcode(ret);
}

void dSMApiWrapper::getActiveDevicesCountInZone(dsid_t dsm_dsid,
                                               uint16_t zone_id,
                                               uint16_t *number_of_devices)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneDeviceCount_only_active(m_dsm_api_handle, dsm_dsid, zone_id,
                                      number_of_devices);
    checkRetcode(ret);
}

void dSMApiWrapper::getInactiveDevicesCountInZone(dsid_t dsm_dsid,
                                                  uint16_t zone_id,
                                                  uint16_t *number_of_devices)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneDeviceCount_only_inactive(m_dsm_api_handle, dsm_dsid, zone_id,
                                      number_of_devices);
    checkRetcode(ret);
}

void dSMApiWrapper::getDevicesCountInZone(dsid_t dsm_dsid,
                                          uint16_t zone_id,
                                          uint16_t *number_of_devices)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneDeviceCount_all(m_dsm_api_handle, dsm_dsid, zone_id,
                                      number_of_devices);
    checkRetcode(ret);
}

void dSMApiWrapper::getActiveDeviceInfoInZone(dsid_t dsm_dsid,
                                             uint16_t req_zone_id,
                                             uint16_t device_index,
                                             uint16_t *device_id,
                                             uint16_t *vendor_id,
                                             uint16_t *product_id,
                                             uint16_t *function_id,
                                             uint16_t *version,
                                             uint16_t *zone_id,
                                             uint8_t *state,
                                             uint8_t *my_click,
                                             uint8_t *locked,
                                             uint8_t *output_mode,
                                             uint8_t *groups,
                                             uint8_t *name,
                                             uint8_t *buttonGroups,
                                             uint8_t *button_zone_membership,
                                             uint32_t *serial_number)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = DeviceInfo_by_index_only_active(m_dsm_api_handle, dsm_dsid,
                                          req_zone_id, device_index, device_id,
                                          vendor_id, product_id, function_id,
                                          version, zone_id, state, my_click,
                                          locked, output_mode, groups,
                                          name, buttonGroups,
                                          button_zone_membership,
                                          serial_number);
    checkRetcode(ret);
}

void dSMApiWrapper::getDeviceInfo(dsid_t dsm_dsid, uint16_t req_device_id,
                                 uint16_t *device_id, uint16_t *vendor_id,
                                 uint16_t *product_id, uint16_t *function_id,
                                 uint16_t *version, uint16_t *zone_id,
                                 uint8_t *state, uint8_t *my_click,
                                 uint8_t *locked, uint8_t *output_mode,
                                 uint8_t *groups, uint8_t *name,
                                 uint8_t *button_groups,
                                 uint8_t *button_zone_membership,
                                 uint32_t *serial_number)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = DeviceInfo_by_device_id(m_dsm_api_handle, dsm_dsid, req_device_id,
                                  device_id, vendor_id, product_id,
                                  function_id, version, zone_id, state,
                                  my_click, locked, output_mode,
                                  groups, name, button_groups,
                                  button_zone_membership, serial_number);
    checkRetcode(ret);
}

void dSMApiWrapper::blinkDevice(dsid_t dsm_dsid, uint16_t device_id)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceActionRequest_action_blink(m_dsm_api_handle, dsm_dsid,
                                               device_id);
    checkRetcode(ret);
}

void dSMApiWrapper::testDeviceTransmissionQuality(dsid_t dsm_dsid,
                                                  uint16_t device_id,
                                                  uint8_t *downstreamQ,
                                                  uint16_t *upstreamQ)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = TestTransmissionQuality_get_sync(m_dsm_api_handle, dsm_dsid,
                                               device_id,
                                               DEFAULT_SYNC_CALL_TIMEOUT,
                                               downstreamQ,
                                               upstreamQ);
    checkRetcode(ret);
}

void dSMApiWrapper::setCommunicationLogging(dsid_t dsm_dsid, uint8_t state)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_set_communication_log_state(m_dsm_api_handle, dsm_dsid, state);
    checkRetcode(ret);
}

void dSMApiWrapper::getCommunicationLogging(dsid_t dsm_dsid, uint8_t* state)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_get_communication_log_state(m_dsm_api_handle, dsm_dsid, state);
    checkRetcode(ret);
}

std::string dSMApiWrapper::error2Message(int err)
{
    // TODO: add all errors
    switch (err)
    {
        case 0:
            return "OK";
        case ERROR_RESPONSE_TIMEOUT:
            return "TIMEOUT";
        default:
            return "Error code " + util::str(err);
    }
}

void dSMApiWrapper::checkRetcode(int ret)
{
    if (ret != 0)
    {
        throw _ExceptionWithCode(error2Message(ret) + ": " + util::str(ret),
                                 ret);
    }
}

dSMApiWrapper::~dSMApiWrapper()
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    if (m_dsm_api_handle)
    {
        DsmApiClose(m_dsm_api_handle);
        DsmApiCleanup(m_dsm_api_handle);
    }
}

}; // namespace

