/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __DSA_DSM_API_WRAPPER_H__
#define __DSA_DSM_API_WRAPPER_H__

#include <digitalSTROM/dsm-api-v2/dsm-api-const.h>
#include <digitalSTROM/dsm-api-v2/dsm-api.h>

#include <boost/thread/mutex.hpp>
#include <string>
#include <list>

#define DSM_API_MAX_DEVICES 63

/// \file dsmapi-wrapper.h
namespace dSA
{

class dSMApiListener
{
public:
    /// dS485-Bus-Level Callbacks
    virtual void busStateCallback(bus_state_t state) = 0;
    virtual void busChangeCallback(dsid_t *id, int flag) = 0;

    /// Device-Level Callbacks
    virtual void deviceAddedCallback(uint8_t error_code, dsid_t src_dsm_dsid,
                                     dsid_t dest_dsm_dsid, uint16_t device_id,
                                     uint16_t zone_id, uint16_t vendor_id,
                                     uint32_t serial_number) = 0;
    virtual void deviceRemovedCallback(uint8_t error_code, dsid_t src_dsm_dsid,
                                       dsid_t dest_dsm_dsid, uint16_t device_id,
                                       uint16_t zone_id, uint16_t vendor_id,
                                       uint32_t serial_number) = 0;
    virtual void deviceStatisticsCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t offset, uint16_t value) = 0;
    virtual void deviceActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr, uint8_t click_type) = 0;
    virtual void deviceLocalActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t state) = 0;
    virtual void deviceConfigCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value) = 0;
    virtual void deviceTestTransmissionQualityCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t downstream_quality, uint16_t upstream_quality) = 0;

    /// Communication Callbacks
    virtual void commUpstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type,
            uint8_t quality, uint8_t flags) = 0;
    virtual void commUpstreamLongCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier, uint32_t data,
            uint8_t quality, uint8_t flags) = 0;
    virtual void commDownstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command) = 0;
    virtual void commDownstreamLong0Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command) = 0;
    virtual void commDownstreamLong8Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command, uint8_t data) = 0;
    virtual void commDownstreamLong16Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command, uint16_t data) = 0;
};

/// This class provides a threadsafe access to the dSM API library.
///
/// Each new instance of the class will open a new connection to the dS485
/// daemon. However each instance is threadsafe and can be used from various
/// threads. Currently it will only wrap the functions that we are interested
/// in, more calls will be added on demand.
///
/// Instead of return values the wrapped calls to the lib will throw exceptions.
class dSMApiWrapper
{
public:
    dSMApiWrapper(std::string ds485_daemon_url);
    ~dSMApiWrapper();

    void addListener(dSMApiListener* listener);
    void removeListener(dSMApiListener* listener);

    void getOwnDSID(dsid_t* dsid);
    bus_state_t getBusState();
    int getBusMembers(dsid_t *device_list, int device_list_size);

    /// dSMeter Infos

    void dSMInfo(dsid_t dsm_dsid, uint32_t *hw_version,
                 uint32_t *arm_sw_version, uint32_t *dsp_sw_version,
                 uint16_t *api_version, dsid_t *dsid, uint8_t *name);
    void dSMState(dsid_t dsm_dsid, uint8_t *hardware, uint8_t *communicationdsp,
                  uint8_t *temperature, uint8_t *configuration, uint8_t *ds485bus,
                  uint8_t *powerline);
    void getApartmentId(dsid_t dsm_dsid, uint8_t *id);
    void getConfigurationId(dsid_t dsm_dsid, uint32_t *hash, uint32_t *configurationcount);
    void getDSMtemperature(dsid_t dsm_dsid, uint16_t *temperature);
    void getZoneCount(dsid_t dsm_dsid, uint8_t *count);
    void getZoneInfo(dsid_t dsm_dsid, uint8_t zone_index, uint16_t *zone_id,
                      uint8_t *virtual_zone_id, uint8_t *number_of_groups,
                      uint8_t *name);
    void getActiveDevicesCountInZone(dsid_t dsm_dsid, uint16_t zone_id,
                                     uint16_t *number_of_devices);
    void getInactiveDevicesCountInZone(dsid_t dsm_dsid, uint16_t zone_id,
                                       uint16_t *number_of_devices);
    void getDevicesCountInZone(dsid_t dsm_dsid, uint16_t zone_id,
                               uint16_t *number_of_devices);
    void getActiveDeviceInfoInZone(dsid_t dsm_dsid, uint16_t req_zone_id,
                             uint16_t device_index, uint16_t *device_id,
                             uint16_t *vendor_id, uint16_t *product_id,
                             uint16_t *function_id, uint16_t *version,
                             uint16_t *zone_id, uint8_t *state,
                             uint8_t *my_click, uint8_t *locked,
                             uint8_t *output_mode, uint8_t *groups,
                             uint8_t *name, uint8_t *buttonGroups,
                             uint8_t *button_zone_membership,
                             uint32_t *serial_number);
    void getDeviceInfo(dsid_t dsm_dsid, uint16_t req_device_id,
                       uint16_t *device_id, uint16_t *vendor_id,
                       uint16_t *product_id, uint16_t *function_id,
                       uint16_t *version, uint16_t *zone_id, uint8_t *state,
                       uint8_t *my_click, uint8_t *locked,
                       uint8_t *output_mode, uint8_t *groups,
                       uint8_t *name, uint8_t *button_groups,
                       uint8_t *button_zone_membership,
                       uint32_t *serial_number);

    /// dSMeter Actions
    void restartRegistration(dsid_t dsm_dsid);
    void setCommunicationLogging(dsid_t dsm_dsid, uint8_t state);
    void getCommunicationLogging(dsid_t dsm_dsid, uint8_t* state);

    /// Device Actions
    void blinkDevice(dsid_t dsm_dsid, uint16_t device_id);
    void testDeviceTransmissionQuality(dsid_t dsm_dsid, uint16_t device_id,
                                       uint8_t *downstreamQ,
                                       uint16_t *upstreamQ);
    static std::string error2Message(int err);

private:
    boost::mutex m_libdsm_mutex;
    boost::mutex m_listener_mutex;
    std::string m_daemon_url;
    DsmApiHandle_t m_dsm_api_handle;
    std::list<dSMApiListener*> m_callback_listeners;

    void checkRetcode(int ret);

    static void busStateCallback(void *arg, bus_state_t state);
    void notifyBusState(bus_state_t state);

    static void busChangeCallback(void *arg, dsid_t *id, int flag);
    void notifyBusChange(dsid_t *id, int flag);

    static void deviceAddedCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
            uint16_t device_id, uint16_t zone_id,
            uint16_t vendor_id, uint32_t serial_number);
    void notifyDeviceAdded(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint16_t vendor_id, uint32_t serial_number);

    static void deviceRemovedCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint16_t vendor_id, uint32_t serial_number);
    void notifyDeviceRemoved(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint16_t vendor_id, uint32_t serial_number);

    static void deviceStatisticsCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t offset, uint16_t value);
    void notifyDeviceStatistics(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t offset, uint16_t value);

    static void deviceActionCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr, uint8_t click_type);
    void notifyDeviceActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr, uint8_t click_type);

    static void deviceLocalActionCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t state);
    void notifyDeviceLocalActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t state);

    static void deviceConfigCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value);
    void notifyDeviceConfigCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value);

    static void deviceTestTransmissionQualityCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t downstream_quality, uint16_t upstream_quality);
    void notifyDeviceTestTransmissionQualityCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t downstream_quality, uint16_t upstream_quality);

    static void commUpstreamShortCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type,
            uint8_t quality, uint8_t flags);
    void notifyCommUpstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type,
            uint8_t quality, uint8_t flags);

    static void commUpstreamLongCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier, uint32_t data,
            uint8_t quality, uint8_t flags);
    void notifyCommUpstreamLongCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier, uint32_t data,
            uint8_t quality, uint8_t flags);

    static void commDownstreamShortCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);
    void notifyCommDownstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);

    static void commDownstreamLong0Callback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);
    void notifyCommDownstreamLong0Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);

    static void commDownstreamLong8Callback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command, uint8_t data);
    void notifyCommDownstreamLong8Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command, uint8_t data);

    static void commDownstreamLong16Callback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command, uint16_t data);
    void notifyCommDownstreamLong16Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command, uint16_t data);
};

}; // namespace

#endif//__DSA_DSM_API_WRAPPER_H__

