/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <boost/shared_ptr.hpp>

#include "json_hardware_web_handler.h"
#include "storage.h"
#include "ds-assistant.h"
#include "exceptions.h"

#define DEFAULT_CHECKS_PER_DEVICE 1

namespace dSA
{

JSONHardwareWebHandler::JSONHardwareWebHandler(struct mg_connection* connection,
                               const struct mg_request_info* request_info) :
                               JSONWebHandler(connection, request_info) {}

std::string JSONHardwareWebHandler::get()
{
    std::ostringstream result;

    Storage *storage = Storage::getInstance();

    boost::shared_ptr<std::list<dsid_t> > dsm_list = storage->getDSMList();
    json_object *ret = json_ok();
    json_object *data = json_object_new_object();
    json_object *dsms = json_object_new_array();
    json_object_object_add(data, "dsm", dsms);

    while (!dsm_list->empty())
    {
        boost::shared_ptr<const dSM> dsm = storage->getDSM(dsm_list->front());
        dsm_list->pop_front();

        if (dsm == NULL)
        {
            continue;
        }

        json_object *jdsm = json_object_new_object();

        json_object_object_add(jdsm, "dsid",
                json_object_new_string(util::str(dsm->getDSID()).c_str()));
        json_object_object_add(jdsm, "hardware",
                json_object_new_int(dsm->getHWVersion()));
        json_object_object_add(jdsm, "arm",
                json_object_new_int(dsm->getARMSwVersion()));
        json_object_object_add(jdsm, "dsp",
                json_object_new_int(dsm->getDSPSwVersion()));
        json_object_object_add(jdsm, "api",
                json_object_new_int(dsm->getAPIVersion()));
        json_object_object_add(jdsm, "name",
                json_object_new_string(dsm->getName().c_str()));

        boost::shared_ptr<std::list<boost::shared_ptr<const Device> > > devs =
            storage->getDevices(dsm->getDSID());

        json_object *devices = json_object_new_array();

        while (!devs->empty())
        {
            boost::shared_ptr<const Device> device = devs->front();
            devs->pop_front();

            if (device == NULL)
            {
                continue;
            }

            json_object *jdevice = json_object_new_object();
            json_object_object_add(jdevice, "dsid",
                json_object_new_string(util::str(device->getDSID()).c_str()));

            json_object_object_add(jdevice, "functionId",
                    json_object_new_int(device->getFunctionId()));

            json_object_object_add(jdevice, "productId",
                    json_object_new_int(device->getProductId()));

            json_object_object_add(jdevice, "version",
                    json_object_new_int(device->getVersion()));

            json_object_object_add(jdevice, "outputMode",
                    json_object_new_int(device->getOutputMode()));

            json_object_object_add(jdevice, "zoneId",
                    json_object_new_int(device->getZoneId()));

            json_object_object_add(jdevice, "serialNumber",
                    json_object_new_int(device->getSerialNumber()));

            json_object_array_add(devices, jdevice);
        }

        json_object_object_add(jdsm, "devices", devices);

        json_object_array_add(dsms, jdsm);
    }

    json_object_object_add(ret, "data", data);

    result << json_object_to_json_string(ret);

    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::testDeviceTransmissionQuality()
{
    std::ostringstream result;

    int checksPerDevice = getOptionalIntParam(m_parameters["checksPerDevice"],
                                              DEFAULT_CHECKS_PER_DEVICE,
                                              "checksPerDevice");

    if (checksPerDevice < 1)
    {
        throw _Exception("parameter " + util::str(checksPerDevice) +
                         " is invalid");
    }

    dsid_t dsm_dsid;

    std::string dsid = m_parameters["dsm"];
    if (!dsid.empty())
    {
        dsm_dsid = util::str2dsid(dsid);
        boost::shared_ptr<const dSM> dsm =
                                    Storage::getInstance()->getDSM(dsm_dsid);
        if (dsm == NULL)
        {
            throw _Exception("dSM " + util::str(dsm_dsid) + " not found");
        }
    }
    else
    {
        SetNullId(dsm_dsid);
    }

    dS_Assistant *dsa = dS_Assistant::getInstance();

    boost::shared_ptr<TestTranmissionQualityTask> task(
            new TestTranmissionQualityTask(checksPerDevice, dsm_dsid));

    unsigned long task_id = dsa->addLongTask(task);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    json_object_object_add(res, "task",
            json_object_new_string(util::str(task_id).c_str()));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

void JSONHardwareWebHandler::process()
{
    try
    {
        std::string result;
        std::string header = http_ok();

        if (m_command == "get")
        {
            result = get();
        }
        else if (m_command == "testDeviceTransmissionQuality")
        {
            result = testDeviceTransmissionQuality();
        }
        else
        {
            failure("unknown function called");
            return;
        }

        mg_write(m_connection, header.c_str(), header.length());
        mg_write(m_connection, result.c_str(), result.length());
    }
    catch (Exception ex)
    {
        failure(ex.getMessage().c_str());
    }
    catch (...)
    {
        failure("exception raised while processing call");
    }
}

} // namespace

