/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#include <digitalSTROM/dsm-api-v2/dsm-api.h>
#include <boost/pointer_cast.hpp>
#include <errno.h>

#include "storage.h"
#include "exceptions.h"
#include "util.h"

namespace dSA
{

Storage* Storage::m_instance = NULL;

Storage* Storage::getInstance()
{
    if (m_instance == NULL)
    {
        m_instance = new Storage();
        if (m_instance == NULL)
        {
            throw _Exception("Could not create Storage class");
        }
    }

    return m_instance;
}

Storage::Storage()
{
    m_unknown.clear();
    m_dsms.clear();
}

void Storage::add(boost::shared_ptr<HWBase> base)
{
    switch (base->getDeviceType())
    {
        case dSMeter:
            {
                boost::shared_ptr<dSM> dsm(
                            boost::dynamic_pointer_cast<dSM>(base)->clone());
                addDSM(dsm);
            }
            break;
        case dSDevice:
            {
                boost::shared_ptr<Device> device(
                    boost::dynamic_pointer_cast<Device>(base)->clone());
                addDevice(device);
            }
            break;
        case Unknown:
        default:
            {
                removeDevice(base->getDSID());
                boost::recursive_mutex::scoped_lock lock(m_mutex);
                m_unknown.push_back(boost::shared_ptr<HWBase>(base->clone()));
            }
           break;
    }
}

boost::shared_ptr<const dSM> Storage::getDSM(dsid_t dsid) const
{
    boost::shared_ptr<Storage::dSMStorage> s = getDSMStorage(dsid);
    if (s == NULL)
    {
        return boost::shared_ptr<const dSM>();
    }

    boost::recursive_mutex::scoped_lock lock(m_mutex);
    return s->getDSM();
}

boost::shared_ptr<std::list<dsid_t> > Storage::getDSMList() const
{
    boost::shared_ptr<std::list<dsid_t> > dsm_list(new std::list<dsid_t>);

    boost::recursive_mutex::scoped_lock lock(m_mutex);
    for (std::list<boost::shared_ptr<dSMStorage> >::const_iterator it =
            m_dsms.begin();
        it != m_dsms.end(); it++)
    {
        boost::shared_ptr<dSMStorage> s = *it;
        dsm_list->push_back(s->getDSID());
    }

    return dsm_list;
}

boost::shared_ptr<std::list<boost::shared_ptr<const Device> > >
                                Storage::getDevices(dsid_t dsm_dsid) const
{
    boost::shared_ptr<Storage::dSMStorage> s = getDSMStorage(dsm_dsid);
    if (s == NULL)
    {
        throw _Exception("dSM with id " + util::str(dsm_dsid) + " is not known"
                         " to us");
    }

    boost::recursive_mutex::scoped_lock lock(m_mutex);
    return s->getDevices();
}

boost::shared_ptr<std::list<boost::shared_ptr<const Device> > >
                                Storage::getDevices() const
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    boost::shared_ptr<std::list<dsid_t> > dsm_list = getDSMList();

    boost::shared_ptr<std::list<boost::shared_ptr<const Device> > > dev_list(
        new std::list<boost::shared_ptr<const Device> >);;
    while (!dsm_list->empty())
    {
        boost::shared_ptr<const dSM> dsm = getDSM(dsm_list->front());
        dsm_list->pop_front();

        boost::shared_ptr<std::list<boost::shared_ptr<const Device> > > devs =
            getDevices(dsm->getDSID());

        while (!devs->empty())
        {
            dev_list->push_back(devs->front());
            devs->pop_front();
        }
    }

    return dev_list;;
}
void Storage::addDSM(boost::shared_ptr<dSM> dsm)
{
    boost::shared_ptr<Storage::dSMStorage> s = getDSMStorage(dsm->getDSID());
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    if (s == NULL)
    {
        m_dsms.push_back(
          boost::shared_ptr<Storage::dSMStorage>(new Storage::dSMStorage(dsm)));
    }
    else
    {
        s->update(dsm);
    }
}

boost::shared_ptr<const HWBase> Storage::get(dsid_t dsid) const
{
    if (DsmApiIsdSM(dsid))
    {
        boost::shared_ptr<const dSM> dsm = getDSM(dsid);
        if (dsm == NULL)
        {
            return boost::shared_ptr<const HWBase>();
        }

        boost::shared_ptr<const HWBase> h =
            boost::dynamic_pointer_cast<const HWBase>(dsm);
        return h;
    }
    else
    {
        // since the dsid does not belong to a dsm it can be either an
        // unknown device or a "klemme"
        boost::shared_ptr<const HWBase> h = getUnknown(dsid);
        if (h != NULL)
        {
            return h;
        }

        boost::recursive_mutex::scoped_lock lock(m_mutex);
        for (std::list<boost::shared_ptr<dSMStorage> >::const_iterator it =
                m_dsms.begin();
            it != m_dsms.end(); it++)
        {
            boost::shared_ptr<dSMStorage> s = *it;
            boost::shared_ptr<const Device> d = s->getDevice(dsid);
            if (d != NULL)
            {
                h = boost::dynamic_pointer_cast<const HWBase>(d);
                return h;
            }
        }
    }

    return boost::shared_ptr<const HWBase>();
}

boost::shared_ptr<const Device> Storage::getDevice(dsid_t dsid,
                                                   dsid_t dsm_dsid) const
{
    boost::shared_ptr<dSMStorage> s = getDSMStorage(dsm_dsid);
    if (s == NULL)
    {
        return boost::shared_ptr<const Device>();
    }

    boost::recursive_mutex::scoped_lock lock(m_mutex);
    return s->getDevice(dsid);
}

boost::shared_ptr<const Device> Storage::getDevice(uint16_t device_id,
                                                   dsid_t dsm_dsid) const
{
    boost::shared_ptr<dSMStorage> s = getDSMStorage(dsm_dsid);
    if (s == NULL)
    {
        return boost::shared_ptr<const Device>();
    }

    boost::recursive_mutex::scoped_lock lock(m_mutex);
    return s->getDevice(device_id);
}

void Storage::addDevice(boost::shared_ptr<Device> device)
{
    // will throw exception if dsm is not found
    boost::shared_ptr<dSMStorage> s = getDSMStorage(device->getDSMdSID());
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    if (s == NULL)
    {
        throw _Exception("dSM with id " + util::str(device->getDSMdSID()) +
                         " not found in storage");
    }

    s->addDevice(device);
}

/// \todo throw exception or return NULL when something was not found?
boost::shared_ptr<Storage::dSMStorage> Storage::getDSMStorage(dsid_t dsm_dsid) const
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    for (std::list<boost::shared_ptr<Storage::dSMStorage> >::const_iterator it =
            m_dsms.begin();
         it != m_dsms.end(); it++)
    {
        boost::shared_ptr<Storage::dSMStorage> s = *it;
        dsid_t id = s->getDSM()->getDSID();

        if (IsEqualId(dsm_dsid, id))
        {
            return s;
        }
    }

    return boost::shared_ptr<Storage::dSMStorage>();
}

void Storage::removeDSMStorage(dsid_t dsm_dsid)
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    for (std::list<boost::shared_ptr<dSMStorage> >::iterator it =
            m_dsms.begin();
         it != m_dsms.end(); it++)
    {
        boost::shared_ptr<dSMStorage> s = *it;
        dsid_t id = s->getDSM()->getDSID();

        if (IsEqualId(dsm_dsid, id))
        {
            m_dsms.erase(it);
            return;
        }
    }
}

bool Storage::removeUnknown(dsid_t dsid)
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    for (std::list<boost::shared_ptr<HWBase> >::iterator it =
            m_unknown.begin();
         it != m_unknown.end(); it++)
    {
        boost::shared_ptr<HWBase> h = *it;
        dsid_t id = h->getDSID();

        if (IsEqualId(dsid, id))
        {
            m_unknown.erase(it);
            return true;
        }
    }

    return false;
}

boost::shared_ptr<const HWBase> Storage::getUnknown(dsid_t dsid) const
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    for (std::list<boost::shared_ptr<HWBase> >::const_iterator it =
            m_unknown.begin();
         it != m_unknown.end(); it++)
    {
        boost::shared_ptr<HWBase> h = *it;
        dsid_t id = h->getDSID();

        if (IsEqualId(dsid, id))
        {
            return h;
        }
    }

    return boost::shared_ptr<const HWBase>();
}

void Storage::removeDevice(dsid_t dsid, dsid_t dsm_dsid)
{
    boost::shared_ptr<dSMStorage> s = getDSMStorage(dsm_dsid);
    if (s != NULL) {
        s->removeDevice(dsid);
    }
}

void Storage::removeDevice(dsid_t dsid)
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    for (std::list<boost::shared_ptr<dSMStorage> >::const_iterator it =
            m_dsms.begin();
        it != m_dsms.end(); it++)
    {
        boost::shared_ptr<dSMStorage> s = *it;
        if (s->removeDevice(dsid))
        {
            return;
        }
    }
}


void Storage::remove(dsid_t dsid)
{
    if (DsmApiIsdSM(dsid))
    {
        removeDSMStorage(dsid);
    }
    else
    {
        // since the dsid does not belong to a dsm it can be either an
        // unknown device or a "klemme"
        if (removeUnknown(dsid))
        {
            return;
        }

        removeDevice(dsid);
    }
}

Storage::dSMStorage::dSMStorage(boost::shared_ptr<dSM> dsm) : m_dsm(dsm) {}

void Storage::dSMStorage::update(boost::shared_ptr<dSM> dsm)
{
    dsid_t old = m_dsm->getDSID();
    dsid_t id = dsm->getDSID();
    if (!IsEqualId(old, id))
    {
        throw _Exception("Can't update dSM in storage, dSID does not match!");
    }

    m_dsm = dsm;
}

boost::shared_ptr<const dSM> Storage::dSMStorage::getDSM() const
{
    return m_dsm;
}

/// \todo throw exception or return NULL when something was not found?
boost::shared_ptr<const Device> Storage::dSMStorage::getDevice(dsid_t dsid) const
{
    for (std::list<boost::shared_ptr<Device> >::const_iterator it = m_devices.begin();
         it != m_devices.end();
         it++)
    {
        dsid_t dev = (*it)->getDSID();
        if (IsEqualId(dev, dsid))
        {
            boost::shared_ptr<Device> d = *it;
            return d;
        }
    }

    return boost::shared_ptr<const Device>();
}

/// \todo throw exception or return NULL when something was not found?
boost::shared_ptr<const Device> Storage::dSMStorage::getDevice(uint16_t device_id) const
{
    for (std::list<boost::shared_ptr<Device> >::const_iterator it = m_devices.begin();
            it != m_devices.end();
            it++)
    {
        uint16_t devid = (*it)->getId();
        if (devid == device_id) {
            boost::shared_ptr<Device> d = *it;
            return d;
        }
    }

    return boost::shared_ptr<const Device>();
}

void Storage::dSMStorage::addDevice(boost::shared_ptr<Device> device)
{
    removeDevice(device->getDSID());
    m_devices.push_back(device);
}

bool Storage::dSMStorage::removeDevice(dsid_t dsid)
{
    for (std::list<boost::shared_ptr<Device> >::iterator it =
            m_devices.begin();
         it != m_devices.end(); it++)
    {
        boost::shared_ptr<Device> d = *it;
        dsid_t id = d->getDSID();
        if (IsEqualId(dsid, id))
        {
            m_devices.erase(it);
            return true;
        }
    }
    return false;
}

const dsid_t Storage::dSMStorage::getDSID() const
{
    return m_dsm->getDSID();
}

boost::shared_ptr<std::list<boost::shared_ptr<const Device> > >
    Storage::dSMStorage::getDevices() const
{
    boost::shared_ptr<std::list<boost::shared_ptr<const Device> > > devices(
                         new std::list<boost::shared_ptr<const Device> >);

    for (std::list<boost::shared_ptr<Device> >::const_iterator it =
            m_devices.begin();
         it != m_devices.end(); it++)
    {
        boost::shared_ptr<Device> d = *it;
        devices->push_back(d);
    }

    return devices;
}


} // namespace
