/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __STORAGE_H__
#define __STORAGE_H__

#include <digitalSTROM/ds.h>
#include <boost/shared_ptr.hpp>
#include <boost/thread/recursive_mutex.hpp>
#include <list>

#include "hwdesc.h"

/// \file storage.h
namespace dSA
{

/// This class stores information about all discovered digitalSTROM devices.
///
/// We have a number of dS-Meters, each dS-Meter has a number of associated
/// devices or "Klemmen". The "Klemmen" are stored within each dSM and are
/// managed by the dSM class, however the Storage class also offers convenience
/// functions for working with devices.
/// We also have a number of "other" devices on the dS485
/// bus; for example a dSS will also show up in the bus members list. We can't
/// really do anything with those, but we will simply track their presence and
/// show them in our list.
///
/// This class is thread safe.
class Storage
{
public:
    static Storage *getInstance();

    void add(boost::shared_ptr<HWBase> base);
    void addDSM(boost::shared_ptr<dSM> dsm);
    void addDevice(boost::shared_ptr<Device> device);

    void remove(dsid_t dsid);
    void removeDSM(dsid_t dsid);
    void removeDevice(dsid_t);
    void removeDevice(dsid_t, dsid_t dsm_dsid);

    boost::shared_ptr<const HWBase> get(dsid_t dsid) const;
    boost::shared_ptr<const dSM> getDSM(dsid_t dsid) const;
    boost::shared_ptr<const Device> getDevice(dsid_t dsid, dsid_t dsm_dsid) const;
    boost::shared_ptr<const Device> getDevice(uint16_t deviceid, dsid_t dsm_dsid) const;

    boost::shared_ptr<std::list<dsid_t> > getDSMList() const;
    boost::shared_ptr<std::list<boost::shared_ptr<const Device> > > getDevices()
                                                                    const;
    boost::shared_ptr<std::list<boost::shared_ptr<const Device> > > getDevices(
            dsid_t dsm_dsid) const;


protected:
    /// \brief The dSM Storage class holds information about the dSM and it's
    /// devices.
    ///
    /// This class is not thread safe, access must be locked.
    class dSMStorage
    {
    public:
        dSMStorage(boost::shared_ptr<dSM> dsm);

        /// \brief Retrieve dSID of the dSM
        const dsid_t getDSID() const;

        /// \brief Replace stored information about the dSM.
        ///
        /// Effectively the iformation stored about the dSM will be replaced
        /// by the one specified in the function argument. It is important
        /// that it is still the *same* physical dSM, otherwise the associated
        /// devices will not match. A safecheck on the dSID is performed, but
        /// otherwise the caller has to know what he is doing.
        void update(boost::shared_ptr<dSM> dsm);

        /// \brief Returns the associated dS-Meter.
        boost::shared_ptr<const dSM> getDSM() const;

        /// \brief Returns the device by dSID
        boost::shared_ptr<const Device> getDevice(dsid_t dsid) const;

        /// \brief Returns the device by Address
        boost::shared_ptr<const Device> getDevice(uint16_t device_id) const;

        /// \brief Adds a new device, if a device with the same dSID already
        /// exists the old device will be deleted first.
        void addDevice(boost::shared_ptr<Device> device);

        /// \brief Removes device from the list.
        /// \return true if device was found and removed.
        bool removeDevice(dsid_t dsid);

        boost::shared_ptr<std::list<boost::shared_ptr<const Device> > >
                                                            getDevices() const;


    protected:
        boost::shared_ptr<dSM> m_dsm;
        std::list<boost::shared_ptr<Device> > m_devices;
    };

    Storage();
    boost::shared_ptr<Storage::dSMStorage> getDSMStorage(dsid_t dsm_dsid) const;
    void removeDSMStorage(dsid_t dsm_dsid);
    bool removeUnknown(dsid_t dsid);
    boost::shared_ptr<const HWBase> getUnknown(dsid_t dsid) const;
    std::list<boost::shared_ptr<HWBase> > m_unknown;
    std::list<boost::shared_ptr<dSMStorage> > m_dsms;

private:
    static Storage* m_instance;
    mutable boost::recursive_mutex m_mutex;
};

};

#endif//__STORAGE_H__
