/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __TASKS_H__
#define __TASKS_H__

#include <digitalSTROM/dsm-api-v2/dsm-api.h>
#include <boost/thread/mutex.hpp>
#include <string>
#include <time.h>

namespace dSA
{

class dS_Assistant;

// these values are saved in the database, so make sure to only add
typedef enum
{
    UnknownTask             = 0,
    TestTransmissionQuality = 1,
} task_type;

typedef enum
{
    UnknownStatus       = 0,
    Queued              = 1,
    Running             = 2,
    Finished            = 3,
    Cancelled           = 4
} task_status;

class dSTaskInfo
{
public:
    dSTaskInfo();
    void setDescription(std::string description);
    const std::string getName() const;
    const std::string getDescription() const;
    bool isValid() const;
    void invalidate();
    bool isCancellable() const;

    virtual ~dSTaskInfo();

protected:
    std::string m_name;
    std::string m_description;
    bool m_valid;
    bool m_cancellable;
};

class dSLongTaskInfo : public dSTaskInfo
{
public:
    dSLongTaskInfo();
    void setTaskID(long long int task_id);
    long long int getTaskID() const;
    void setParentTaskID(long long int task_id);
    long long int getParentTaskID() const;
    unsigned char getProgress() const;
    std::string getStatusInfo() const;
    std::string getParameters() const;
    void setTaskStatus(task_status status);
    task_status getTaskStatus() const;
    std::string getTaskStatusText() const;
    virtual task_type getTaskType() const = 0;
    time_t getStartedTime() const;
    void setStartedTime(time_t t);
    time_t getFinishedTime() const;
    void setFinishedTime(time_t t);
    time_t getQueuedTime() const;
    void setQueuedTime(time_t t);

protected:
    long long int m_task_id;
    long long int m_parent_task_id;
    unsigned char m_progress;
    std::string m_status_info;
    task_status m_task_status;
    mutable boost::mutex m_status_mutex;
    std::string m_parameters;
    time_t m_started;
    time_t m_finished;
    time_t m_queued;
};

class dSTask : public dSTaskInfo
{
public:
    dSTask();
    virtual void run(dS_Assistant *) = 0;
};

class dSLongTask : public dSLongTaskInfo
{
public:
    dSLongTask();
    virtual void run(dS_Assistant *) = 0;
};

class GetBusMembersTask : public dSTask
{
public:
    GetBusMembersTask(bool async = true);
    virtual void run(dS_Assistant *);
private:
    bool m_async;
};

class GetDSMInfoTask : public dSTask
{
public:
    GetDSMInfoTask(dsid_t dsid, bool async = true);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    bool m_async;
};

class GetDSMZonesTask : public dSTask
{
public:
    GetDSMZonesTask(dsid_t dsid);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
};

class GetDSMZoneInfoTask : public dSTask
{
public:
    GetDSMZoneInfoTask(dsid_t dsid, uint8_t zone_index);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    uint8_t m_zone_index;
};

class ScanDSMTask : public dSTask
{
public:
    ScanDSMTask(dsid_t dsid, bool async = true);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    bool m_async;
};

class GetActiveDevicesInZoneTask : public dSTask
{
public:
    GetActiveDevicesInZoneTask(dsid_t dsid, uint16_t zone_id);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    uint16_t m_zone_id;
};

class GetDeviceInfoTask : public dSTask
{
public:
    GetDeviceInfoTask(dsid_t dsm_dsid, uint16_t device_id);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    uint16_t m_device_id;

};

class TestTranmissionQualityTask : public dSLongTask
{
public:
    TestTranmissionQualityTask(int checksPerDevice, dsid_t dsm_dsid);
    virtual void run(dS_Assistant *);
    virtual task_type getTaskType() const;
private:
    int m_checksPerDevice;
    dsid_t m_dsm_dsid;
};

}

#endif//__TASKS_H__
